#include <stdio.h>
#include <sys/time.h>

#include <isode/quipu/util.h>
#include <isode/quipu/entry.h>

#include "bulkloadattrnames.h"
#include "configuration.h"

extern	struct	bulkLoadConfigStruct bulkLoadConfig ;



/*
 * From the attributes already contained in the Attribute Sequence that
 * is to be used to add an entry determine the value for the
 * "deleteEntryAfter" attribute. Finally include this attribute in
 * the attribute sequence.
 */

void
appendDeleteEntryAfterToAttrSeq( addEntryAS )
Attr_Sequence	addEntryAS ;
{
Attr_Sequence	currAS 	   = addEntryAS ;
AttributeType	classAttrT = NULLAttrT ;
AttributeValue	classAttrV = NULLAttrV ;
int		periodReqBeforeDeletion, /* in months */
		delEntIndex ;
struct 		timeval  t,  *tp  = &t ;
struct		timezone tz, *tzp = &tz ;
struct	 	tm * tmp ;
UTCtime	 	utc, * utcp = &utc ;
char *		UTCtimecp ;


	if( bulkLoadConfig.checkAttrTypeForDeleteEntry == NULLAttrT )
	    return ;
	
	
	/* Search for the attribute that is to be used to define the value to 
 	   be cross-referenced with the period before deletion statements ..  */
	   
	while( currAS )
	{
	    if( AttrT_cmp( currAS -> attr_type, 
	    		   bulkLoadConfig.checkAttrTypeForDeleteEntry ) == 0 )
	    {
		classAttrT = currAS -> attr_type ;
		break ;
	    }
	    else
	    	currAS = currAS -> attr_link ;
	}


	/* Search for the corresponding period before deletion value .. */

	periodReqBeforeDeletion = CONFIGNOLIMITPERIODBEFOREDEL ; /* default to no limit on period before deletion */

	if( classAttrT != NULLAttrT && 
	    currAS != NULLATTR && currAS -> attr_value != NULLAV )
	{
	    classAttrV = &(currAS -> attr_value -> avseq_av) ; /* assuming only one value for the class !! */
	
	    /* Now find the period required for this class.. */

	    for( delEntIndex = 0 ; delEntIndex < bulkLoadConfig.lastUsedPeriodBeforeDeleteToken ; delEntIndex++ )
	    {
	    	if( AttrV_cmp( classAttrV,
			       bulkLoadConfig.periodBeforeDeletionAttrV[ delEntIndex ] ) == 0 )
		{
		    periodReqBeforeDeletion = bulkLoadConfig.periodBeforeDeletionForAttrV[ delEntIndex ] ;
		    break ;
		}
	    }
	}

	if( periodReqBeforeDeletion != CONFIGNOLIMITPERIODBEFOREDEL )
	{
	    /* Now we must work out the actual UTC time for deletion given the current time */

	    /* Buffer the current time */
	    if( !gettimeofday( tp, tzp) )
	    {
		tmp = localtime( &((time_t)(tp -> tv_sec)) ) ;

		tm2ut( tmp, utcp ) ;

	    	utcp -> ut_mon += periodReqBeforeDeletion ;

		/* Adjust the months/years if necessary.. */
		if( utcp -> ut_mon > 12 )
		{
		    utcp -> ut_year += (int)((utcp -> ut_mon - 1)/ 12) ; /* 24 mths -> 1 yr 12 mths */

		    utcp -> ut_mon %= 12 ;

		    if( utcp -> ut_mon == 0 )
		    	utcp -> ut_mon =  12 ;
		}
		
		UTCtimecp = time2str( utcp, 0 ) ;
#ifdef DEBUG
fprintf( stderr, "deleteEntryAfter time = %s\n", UTCtimecp ) ;
#endif
		if( UTCtimecp != NULLCP )
		    appendAttributeToAttrSeq( BL_DELETE_ENTRY_AFTER, UTCtimecp, addEntryAS ) ;
	    }
	}

	return ;
}
