#include <stdio.h>

#include <isode/quipu/list.h>
#include <isode/quipu/dua.h>

#include "bulkloadattrnames.h"
#include "namelist.h"
#include "referral.h"
#include "util.h"

/*
 * Note: if an alias is followed the parentDN parameter will be updated.
 *
 */

int
listSubtree( parentDN, errorMsgcpp, subordinatespp )
DN   	parentDN ;
char **	errorMsgcpp ;
struct 	subordinate ** subordinatespp ;
{
static 	struct ds_list_arg 	larg;
static 	struct ds_list_result 	lresult, * currListResultp ;
static 	struct DSError 		lerror;
static 	CommonArgs lca = default_common_args;
int	retVal = NOTOK, loopCount = 0 ;
struct	subordinate * currSubordinatep ; 
char *	DNcp ;

	if (rebind() != OK)
	    return NOTOK;

	larg.lsa_common = lca; /* struct copy */
	larg.lsa_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	larg.lsa_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	larg.lsa_common.ca_servicecontrol.svc_options  |= SVC_OPT_DONTUSECOPY ;
	larg.lsa_common.ca_servicecontrol.svc_options  |= SVC_OPT_DONTDEREFERENCEALIAS ;

	larg.lsa_object = dn_cpy( parentDN ) ;

	if( (DNcp = dn2pstr( parentDN )) != NULLCP )
	{
	    resetprint( "BULKLOAD is currently listing\n%s\n", DNcp ) ;
	    free( DNcp ) ;
	}
	
	do {

	    startUnbindTimer() ;
	    retVal = de_list(&larg, &lerror, &lresult);
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( retVal, &lerror ) == TRUE ) ;


	if( retVal != OK )
	{
	    retVal = NOTOK ;
	    *errorMsgcpp = error2pstr( &lerror ) ; /* ds_error_free() called below */
	}
	else
	{
	    currListResultp  = &lresult ; 
	    currSubordinatep = NULLSUBORD ;

	    while( currListResultp != ((struct ds_list_result *)NULL ) && retVal == OK )
	    {
	    	dn_free( currListResultp->lsr_object ) ;

		if( currListResultp->lsr_poq.poq_limitproblem != LSR_NOLIMITPROBLEM )
		{
		    *errorMsgcpp = strdup( "Limit problem list aborted" ) ;		    
		    retVal = NOTOK ;
		}
		
		crefs_free( currListResultp->lsr_poq.poq_cref ) ;

		if( currSubordinatep == NULLSUBORD )
		{
		    *subordinatespp  = lresult.lsr_subordinates ;		    
		    currSubordinatep = *subordinatespp ;
		}
		else
		{
		    /* Move the subordinates pointer to the last node held so far.. */
		    for( ; ; )
		    {
		    	if( currSubordinatep -> sub_next != NULLSUBORD )
			    currSubordinatep = currSubordinatep -> sub_next ;
			else
			    break ;
		    }

		    /* Link in the new subordinates.. */
		    currSubordinatep -> sub_next = currListResultp -> lsr_subordinates ;
		}

	    	currListResultp = currListResultp -> lsr_next ;
	    }
	}

	dn_free( larg.lsa_object ) ;

	return( retVal ) ;
}


void
free_subordinatepp( subordinatespp )
struct	subordinate ** subordinatespp ;
{
struct	subordinate * currsubp, *prevsubp ;

	if( subordinatespp == NULL )
	    return ;

	currsubp = *subordinatespp ;
	
	while( currsubp != NULLSUBORD )
	{
	    if( currsubp -> sub_rdn != NULLRDN )
		rdn_free( currsubp -> sub_rdn ) ;

	    prevsubp = currsubp ;
	    currsubp = currsubp -> sub_next ;
	    free( (char *)prevsubp ) ; /* ..return the node itself */
	}

	*subordinatespp = NULLSUBORD ;

	return ;
}


/*
 * Descend the tree depth first applying the specified operation to leaf and 
 * non-leaf nodes as required.
 *
 * Note: the sub-ordinates are stored in the next level down from the parent and
 * so level 0 will not have any.
 */

#define	MAXLEVELS	30

int
applyOperationToSubTreeNodes( outfp, parentDN, nodeOperation, applyToJustLeafNodes, nodeParamcp, nodeParamAttrT, nodeParamAttrV )
FILE *	outfp ; /* May be NULL */
DN	parentDN ;
int	(*nodeOperation)() ;
int	applyToJustLeafNodes ; /* if FALSE then leaf AND non-leaf nodes */
char *	nodeParamcp ;
AttributeType	nodeParamAttrT ;
AttributeValue	nodeParamAttrV ;
{
int	retVal = OK, maxlevel, index, loopExit = FALSE, firstCall = TRUE ;
char *	errorMsgcp, * DNcp ;
AttributeType	cnat = NULLAttrT, uidat = NULLAttrT, dataSourceat = NULLAttrT, bulkSubDirListat = NULLAttrT ;
struct	subordinate * subordinatesp, * currSubordp ;
struct	subordinate * levelBaseSubordinates[ MAXLEVELS ] ;
struct	subordinate * levelCurrSubordinates[ MAXLEVELS ] ;
DN	levelParentDN[ MAXLEVELS ], nodeDN ;

	maxlevel = 0 ; /* current deepest level */
	levelParentDN[ maxlevel ] = dn_cpy( parentDN ) ;
	levelBaseSubordinates[ maxlevel ] = levelCurrSubordinates[ maxlevel ] = NULLSUBORD ;

	uidat = str2AttrT( BL_UID ) ; cnat = str2AttrT( BL_COMMON_NAME ) ; 
	dataSourceat = str2AttrT( BL_DATA_SOURCE ) ; bulkSubDirListat = str2AttrT( BL_BULK_SUB_DIR_LIST ) ;

	while( loopExit == FALSE )
	{
	    subordinatesp = NULLSUBORD ;

	    if( (levelCurrSubordinates[ maxlevel ] == NULLSUBORD ||
	    	 (findAttrTypeRdnComp( cnat, levelCurrSubordinates[ maxlevel ] -> sub_rdn ) == NULLRDN &&
	          findAttrTypeRdnComp( uidat,levelCurrSubordinates[ maxlevel ] -> sub_rdn ) == NULLRDN &&
	          findAttrTypeRdnComp( dataSourceat,levelCurrSubordinates[ maxlevel ] -> sub_rdn ) == NULLRDN &&
	          findAttrTypeRdnComp( bulkSubDirListat, levelCurrSubordinates[ maxlevel ] -> sub_rdn ) == NULLRDN) )
		/* ..limit the number of list operations */
	    	&& listSubtree( levelParentDN[ maxlevel ], &errorMsgcp, &subordinatesp ) == NOTOK )
	    {
	    	fprintf( stderr, "%s\n", errorMsgcp ) ;
		if( outfp != ((FILE *)NULL) ) {
		    if( levelParentDN[ maxlevel ] != NULLDN &&
			(DNcp = dn2pstr( levelParentDN[ maxlevel ] )) != NULLCP )
		    {
			fprintf( outfp, "\n# %s\n", DNcp ) ;
			free( DNcp ) ;
		    }
		    fprintf( outfp,  "# %s\n", errorMsgcp ) ;
		}
	    	free( errorMsgcp ) ;
	    	loopExit = TRUE ; retVal = NOTOK ;
	    }
	    else
	    {
	    	if( subordinatesp == NULLSUBORD )
		{
		    if( levelCurrSubordinates[ maxlevel ] == NULLSUBORD )
		    {
		    	/* Initial list did not return any subordinates ! */
		        loopExit = TRUE ; retVal = NOTOK ; 
			break ;
		    }
		
		    if( applyToJustLeafNodes == TRUE )
		    {
		    	/* Apply operation to this leaf node.. */
			if( (retVal = nodeOperation( outfp, levelParentDN[ maxlevel ], nodeParamcp, nodeParamAttrT, nodeParamAttrV, firstCall )) == NOTOK )
		    	{
		    	    loopExit = TRUE ;
			    break ;
			}
		    }

		    firstCall = FALSE ;

		    while( levelCurrSubordinates[ maxlevel ] -> sub_next == NULLSUBORD )
		    {
		    	/* Move up a level after first freeing the current maximum level's subordinate.. */
			
		    	free_subordinatepp( &(levelBaseSubordinates[ maxlevel ]) ) ;
			dn_free( levelParentDN[ maxlevel ] ), levelParentDN[ maxlevel ] = NULLDN ;

		    	--maxlevel ;

		    	if( maxlevel == 0 )
			{
			    loopExit = TRUE ;
			    break ; /* Successful completion ! */
			}
		    }

		    if( loopExit == FALSE )
		    {
		    	levelCurrSubordinates[ maxlevel ] = levelCurrSubordinates[ maxlevel ] -> sub_next ;
		    	/* .. Use the next subordinate at the deepest level */

		    	dn_free( levelParentDN[ maxlevel ] ) ;

		    	/* Generate a new list operation DN.. */
		    	levelParentDN[ maxlevel ] = dn_cpy( levelParentDN[ maxlevel - 1 ] ) ;
 		    	dn_append( levelParentDN[ maxlevel ], dn_comp_new( rdn_cpy( levelCurrSubordinates[ maxlevel ] -> sub_rdn) )) ;			
		    }
		}
		else
		{
		    /* The last list operation has returned sub-ordinates.. */
			
		    if( applyToJustLeafNodes == FALSE )
		    {
			if( (DNcp = dn2pstr( levelParentDN[ maxlevel ] )) != NULLCP )
			{
			    resetprint( "BULKLOAD is currently applying an operation to each of the subordinates of\n%s\n", DNcp ) ;
			    free( DNcp ) ;
			}
		    	
			/* Apply to each subordinate (whether leaf or non-leaf).. */
			for( currSubordp = subordinatesp ; loopExit==FALSE && currSubordp != NULLSUBORD ; currSubordp = currSubordp -> sub_next )
			{
			    nodeDN = dn_cpy( levelParentDN[ maxlevel ] ) ;
			    dn_append( nodeDN, dn_comp_new( rdn_cpy( currSubordp -> sub_rdn ))) ;
			    
		    	    if( (retVal = nodeOperation( outfp, nodeDN, nodeParamcp, nodeParamAttrT, nodeParamAttrV, firstCall )) == NOTOK )
		    	    {
		    	    	loopExit = TRUE ;
			    }

			    firstCall = FALSE ; dn_free( nodeDN ) ;
			}

			if( loopExit == TRUE )
			    break ;
		    }
		    
		    if( maxlevel+1 == MAXLEVELS )
		    {
		    	fprintf( stderr, "FATAL error more that %d levels in the sub-tree\n", MAXLEVELS ) ;
			if( outfp != ((FILE *)NULL))
			    fprintf( outfp,  "# FATAL error more that %d levels in the sub-tree\n", MAXLEVELS ) ;

			loopExit = TRUE ; retVal = NOTOK ;
			break ;
		    }

		    ++maxlevel ;

		    /* Buffer the retured subordinate linked list.. */
		    levelBaseSubordinates[ maxlevel ] = subordinatesp ;
		    levelCurrSubordinates[ maxlevel ] = subordinatesp ;


		    /* Generate a new list operation DN.. */
		    levelParentDN[ maxlevel ] = dn_cpy( levelParentDN[ maxlevel - 1 ] ) ;
 		    dn_append( levelParentDN[ maxlevel ], dn_comp_new( rdn_cpy( levelCurrSubordinates[ maxlevel ] -> sub_rdn) )) ;
		}
	    }
	}

	/* Return any remaining used storage.. */

	/* maxlevel should be 0 but just in case ! */
	for( index = 0 ; index <= maxlevel ; index++ )
	    dn_free( levelParentDN[ index ] ) ;

	for( index = 1 ; index <= maxlevel ; index++ )
	    free_subordinatepp( &(levelBaseSubordinates[ index ]) ) ;
	
	AttrT_free( cnat ) ; AttrT_free( uidatat ) ; AttrT_free( dataSourceat ) ; AttrT_free( bulkSubDirListat ) ;

	return( retVal ) ;
}

